/**
 * Practica Tema 7: CLIENTE TFTP 
 * 
 * Redondo Calleja, Daniel
 * Jimenez Prieto, Roberto
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <unistd.h>

#define PUERTO_TFTP 69
#define TAMANO_BLOQUE 512
// Codigos de operaciones de TFTP
#define CODIGO_RRQ 1
#define CODIGO_WRQ 2
#define CODIGO_DATOS 3
#define CODIGO_ACK 4
#define CODIGO_ERROR 5

// Declaraciones de las funciones
void analizar_argumentos(int argc, char *argv[], char **ip_servidor, char **operacion, char **fichero, int *verbose);
int crear_socket(const char *ip_servidor, struct sockaddr_in *direccion_servidor);
void enviar_lectura(int socket, const struct sockaddr_in *direccion_servidor, const char *fichero, int verbose);
void enviar_escritura(int socket, const struct sockaddr_in *direccion_servidor, const char *fichero, int verbose);

int main(int argc, char *argv[]) {
    char *ip_servidor, *operacion, *fichero;
    int verbose;

    // Analizamos los argumentos de entrada
    analizar_argumentos(argc, argv, &ip_servidor, &operacion, &fichero, &verbose);

    // Creamos el socket para la conexion UDP
    struct sockaddr_in direccion_servidor;
    int socket = crear_socket(ip_servidor, &direccion_servidor);

    // Determinamos si la operacion es de lectura o escritura
    if (strcmp(operacion, "-r") == 0) {
        enviar_lectura(socket, &direccion_servidor, fichero, verbose);
    } else if (strcmp(operacion, "-w") == 0) {
        enviar_escritura(socket, &direccion_servidor, fichero, verbose);
    // Si no es ninguna, error
    } else {
        fprintf(stderr, "Operacion no soportada. Usa -r o -w.\n");
        exit(EXIT_FAILURE);
    }

    close(socket);
    return 0;
}

// Funcion para analizar los argumentos de la linea de comandos
void analizar_argumentos(int argc, char *argv[], char **ip_servidor, char **operacion, char **fichero, int *verbose) {
    if (argc < 4 || argc > 5) {
        fprintf(stderr, "Uso: ./tftp-client ip_servidor {-r|-w} fichero [-v]\n");
        exit(EXIT_FAILURE);
    }

    *ip_servidor = argv[1];        // IP del servidor TFTP
    *operacion = argv[2];          // Operacion: lectura (-r) o escritura (-w)
    *fichero = argv[3];            // Nombre del archivo
    *verbose = 0;                  // Por defecto, modo detallado desactivado

    // Comprobamos si el ultimo argumento es -v
    if (argc == 5 && strcmp(argv[4], "-v") == 0) {
        *verbose = 1;              // Activamos modo detallado
    } else if (argc == 5) {
        fprintf(stderr, "Argumento no reconocido: %s\n", argv[4]);
        exit(EXIT_FAILURE);
    }
}


// Funcion para crear el socket UDP y configurar la direccion del servidor
int crear_socket(const char *ip_servidor, struct sockaddr_in *direccion_servidor) {
    int socket_udp = socket(AF_INET, SOCK_DGRAM, 0); // Creamos socket UDP
    if (socket_udp < 0) {
        fprintf(stderr,"Error al crear el socket.\n");
        exit(EXIT_FAILURE);
    }

    memset(direccion_servidor, 0, sizeof(*direccion_servidor));
    direccion_servidor->sin_family = AF_INET;                  // Familia IPv4
    direccion_servidor->sin_port = htons(PUERTO_TFTP);         // Puerto TFTP
    if (inet_pton(AF_INET, ip_servidor, &direccion_servidor->sin_addr) <= 0) {
        fprintf(stderr, "Direccion IP invalida.\n");
        close(socket_udp);
        exit(EXIT_FAILURE);
    }

    return socket_udp;
}

// Funcion para enviar una solicitud de lectura (RRQ) al servidor
void enviar_lectura(int socket, const struct sockaddr_in *direccion_servidor, const char *fichero, int verbose) {
    unsigned char buffer[TAMANO_BLOQUE + 4];
    unsigned char ack[4];
    int continua = 1;

    FILE *archivo_salida = fopen(fichero, "wb");
    if (!archivo_salida) {
        fprintf(stderr, "Error al abrir el fichero para escribir.\n");
        exit(EXIT_FAILURE);
    }

    // Construimos la solicitud RRQ
    buffer[0] = 0; buffer[1] = CODIGO_RRQ;      // Primero el OPTCODE
    int len = 2;

    strcpy((char *)&buffer[len], fichero);      // Despues el contenido
    len += strlen(fichero) + 1;

    strcpy((char *)&buffer[len], "octet");      // Y por ultimo el modo de transferencia
    len += strlen("octet") + 1;

    // Enviamos la solicitud al servidor
    sendto(socket, buffer, len, 0, (struct sockaddr *)direccion_servidor, sizeof(*direccion_servidor));
    if (verbose) printf("Enviada solicitud de lectura de \"%s\" a servidor tftp (%s)\n", fichero, inet_ntoa(direccion_servidor->sin_addr));

    struct sockaddr_in direccion_origen;
    socklen_t tamano_origen = sizeof(direccion_origen);
    int numero_bloque = 1;

    // Recibimos datos en bloques
    while (continua) {
        int recibido = recvfrom(socket, buffer, sizeof(buffer), 0, (struct sockaddr *)&direccion_origen, &tamano_origen);

        // Si el bloque recibido tiene menos de 4 bytes o OPTCODE == 5, ha ocurrido un error
        if (recibido < 4 || buffer[1] == CODIGO_ERROR) {
            fprintf(stderr, "Error al recibir datos o respuesta del servidor.\n");
            fclose(archivo_salida);
            exit(EXIT_FAILURE);
        }

        if (buffer[1] == CODIGO_DATOS) {
            int bloque_recibido = (buffer[2] << 8) | buffer[3];
            if (verbose) {
                printf("Recibido bloque del servidor tftp.\n");
                if (bloque_recibido == 1)
                    printf("Es el primer bloque (numero de bloque %d)\n", bloque_recibido);
                else
                    printf("Es el bloque numero %d\n", bloque_recibido);
            }

            fwrite(buffer + 4, 1, recibido - 4, archivo_salida);

            // Enviamos ACK
            ack[0] = 0; ack[1] = CODIGO_ACK;            // Escribimos el OPTCODE de ACK
            ack[2] = buffer[2]; ack[3] = buffer[3];     // Escribimos el numero de bloque que corresponde

            sendto(socket, ack, 4, 0, (struct sockaddr *)&direccion_origen, tamano_origen);

            if (verbose) printf("Enviamos el ACK del bloque %d.\n", bloque_recibido);

            // Si detectamos que ha sido el ultimo bloque, salimos del bucle
            if (recibido - 4 < TAMANO_BLOQUE) {
                if (verbose) printf("El bloque %d era el ultimo: cerramos el fichero.\n", bloque_recibido);
                continua = 0;
                continue;
            }

            numero_bloque++;
        }
    }

    fclose(archivo_salida);
}


// Funcion para enviar una solicitud de escritura (WRQ) al servidor
void enviar_escritura(int socket, const struct sockaddr_in *direccion_servidor, const char *fichero, int verbose) {
    unsigned char buffer[TAMANO_BLOQUE + 4];
    unsigned char ack[4];
    int continua = 1;

    FILE *archivo_entrada = fopen(fichero, "rb");
    if (!archivo_entrada) {
        fprintf(stderr, "Error al abrir el fichero para lectura.\n");
        exit(EXIT_FAILURE);
    }

    // Construimos la solicitud WRQ
    buffer[0] = 0; buffer[1] = CODIGO_WRQ;      // Primero el OPTCODE
    int len = 2;

    strcpy((char *)&buffer[len], fichero);      // Despues el contenido
    len += strlen(fichero) + 1;

    strcpy((char *)&buffer[len], "octet");      // Y por ultimo el modo de transferencia
    len += strlen("octet") + 1;

    sendto(socket, buffer, len, 0, (struct sockaddr *)direccion_servidor, sizeof(*direccion_servidor));
    if (verbose) printf("Enviada solicitud de escritura de \"%s\" a servidor tftp (%s)\n", fichero, inet_ntoa(direccion_servidor->sin_addr));

    struct sockaddr_in direccion_origen;
    socklen_t tamano_origen = sizeof(direccion_origen);
    int numero_bloque = 1;

    recvfrom(socket, ack, sizeof(ack), 0, (struct sockaddr *)&direccion_origen, &tamano_origen);

    // Enviamos datos en bloques
    while (continua) {
        size_t leido = fread(buffer + 4, 1, TAMANO_BLOQUE, archivo_entrada);

        // Escribimos el OPTCODE 
        buffer[0] = 0; buffer[1] = CODIGO_DATOS;        

        // Escribimos el numero de bloque correspondiente
        buffer[2] = (numero_bloque >> 8) & 0xFF;        // Parte alta del numero
        buffer[3] = numero_bloque & 0xFF;               // Parte baja del numero

        if (verbose) {
            printf("Enviando bloque al servidor tftp.\n");
            if (numero_bloque == 1)
                printf("Es el primer bloque (numero de bloque %d)\n", numero_bloque);
            else
                printf("Es el bloque numero %d\n", numero_bloque);
        }

        sendto(socket, buffer, leido + 4, 0, (struct sockaddr *)&direccion_origen, tamano_origen);

        recvfrom(socket, ack, sizeof(ack), 0, (struct sockaddr *)&direccion_origen, &tamano_origen);

        // Si el OPTCODE del ACK es erroneo o directamente no es ACK, error
        if (ack[1] != CODIGO_ACK || ((ack[2] << 8) | ack[3]) != numero_bloque) {
            fprintf(stderr, "Error de ACK.\n");
            fclose(archivo_entrada);
            exit(EXIT_FAILURE);
        }

        if (verbose) printf("Recibimos el ACK del bloque %d.\n", numero_bloque);

        // Si detectamos que ha sido el ultimo bloque, salimos del bucle
        if (leido < TAMANO_BLOQUE) {
            if (verbose) printf("El bloque %d era el ultimo: cerramos el fichero.\n", numero_bloque);
            continua = 0;
            continue;
        }

        numero_bloque++;
    }

    fclose(archivo_entrada);
}


